import logging
import os
from typing import Iterable

from odp.tabular_v2.big.big import BigCol


class LocalBigCol(BigCol):
    def __init__(self, root: str):
        super().__init__()
        self.root = root
        os.makedirs(self.root, exist_ok=True)

    def fetch(self, big_id: str) -> bytes:
        logging.info("downloading %s", big_id)
        with open(f"{self.root}/{big_id}.big", "rb") as f:
            return f.read()

    def upload(self, big_id: str, data: Iterable[bytes]):
        logging.info("uploading %s", big_id)
        with open(f"{self.root}/{big_id}.big", "wb") as f:
            for d in data:
                f.write(d)



from typing import Dict, Optional
from uuid import UUID

from pydantic import BaseModel, Field

from odp.dto.validators import validate_resource_name


class MetadataDto(BaseModel):
    """Resource manifest metadata"""

    name: str
    """Resource name. Must consist of alphanumeric characters, dashes or underscores and must start
    with an alphanumeric character"""

    display_name: Optional[str] = None
    """Human-friendly name"""

    description: Optional[str] = None
    """Resource description"""

    uuid: Optional[UUID] = None
    """System-assigned unique identifier"""

    labels: Dict = Field(default_factory=dict)
    """Resource labels"""

    owner: Optional[UUID] = None

    _validate_name = validate_resource_name("name")



import io
import logging
from typing import Dict, Iterator, List, Union

import pandas as pd
import pyarrow as pa
from pyarrow.lib import ArrowInvalid

from odp.tabular_v2 import big, bsquare
from odp.tabular_v2.client import Table
from odp.tabular_v2.client.validation import validate_data_against_schema
from odp.tabular_v2.util import exp


class Transaction:
    def __init__(self, table: Table, tx_id: str, buf: big.Buffer):
        if not tx_id:
            raise ValueError("tx_id must not be empty")
        self._table = table
        self._id = tx_id
        self._buf: List[Union[pa.RecordBatch, List[Dict]]] = []
        self._buf_rows = 0
        self._big_buf = buf

    def select(self, query: Union[exp.Op, str, None] = None) -> Iterator[Dict]:
        for row in self._table.select(query).rows():
            yield row

    def replace(self, query: Union[exp.Op, str, None] = None, vars: Union[Dict, List, None] = None) -> Iterator[Dict]:
        """perform a two-step replace:
        rows that don't match the query are kept.
        rows that match are removed and sent to the caller.
        the caller might insert them again or do something else.
        """
        if query is None:
            raise ValueError("For your own safety, please provide a query like 1==1")
        assert self._buf_rows == 0  # FIXME: handle buffered data in replace/select
        for row in self._table._query_cursor(type="replace", query=query, vars=vars, tx_id=self._id).rows():
            yield row

    def delete(self, query: Union[exp.Op, str, None] = None) -> int:
        ct = 0
        for _ in self.replace(query):
            ct += 1
        return ct

    def flush(self):
        logging.info("flushing to stage %s", self._id)
        if len(self._buf) == 0:
            return

        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, self._table._inner_schema)

        for b in self._buf:
            if isinstance(b, list):
                b = pa.RecordBatch.from_pylist(b, schema=self._table._outer_schema)
            df = b.to_pandas()
            df = bsquare.encode(df, self._table._outer_schema)
            df = self._big_buf.encode(df)

            try:
                w.write_batch(pa.RecordBatch.from_pandas(df, schema=self._table._inner_schema))
            except ArrowInvalid as e:
                raise ValueError("Invalid arrow format") from e
        w.close()
        self._table._client._request(
            path="/api/table/v2/insert",
            params={
                "table_id": self._table._id,
                "tx_id": self._id,
            },
            data=buf.getvalue(),
        ).json()
        self._buf = []
        self._buf_rows = 0

    def insert(self, data: Union[Dict, List[Dict], pa.RecordBatch, pd.DataFrame]):
        """queue data to be inserted on flush()"""
        if isinstance(data, dict):
            data = [data]

        validate_data_against_schema(data, self._table._outer_schema)
        if isinstance(data, list):
            # we expand the last list if it's already a list
            last = self._buf[-1] if self._buf else None
            if last and isinstance(last, list):
                last.extend(data)
            else:
                self._buf.append(data)
            self._buf_rows += len(data)
        elif isinstance(data, pd.DataFrame):
            data = pa.RecordBatch.from_pandas(data, schema=self._table._outer_schema)
            self._buf.append(data)
            self._buf_rows += data.num_rows
        elif isinstance(data, pa.RecordBatch):
            self._buf.append(data)
            self._buf_rows += data.num_rows
        else:
            raise ValueError(f"unexpected type {type(data)}")

        if self._buf_rows >= 10_000:
            self.flush()



import base64
import logging
import os
import threading
from time import time
from typing import Optional

from .util import size2human


class Cache:
    class Entry:
        def __init__(self, key: str, cache: "Cache"):
            self.key = key
            self.lock = threading.Lock()
            self.cache = cache
            self.filename = base64.b64encode(key.encode()).decode()
            self.size = 0

        def set(self, value: bytes) -> bool:
            if len(value) > self.cache.max_entry_size:
                return False

            self.cache._make_space(len(value))
            self.cache.tot_bytes -= self.size  # if replacing, this will be non-zero
            self.size = len(value)
            self.cache.tot_bytes += self.size

            with open(self.cache.root_folder + "/" + self.filename, "wb") as f:
                f.write(value)

        def exists(self) -> bool:
            filename = self.cache.root_folder + "/" + self.filename
            return os.path.exists(filename)

        def age(self) -> float:
            return time() - os.path.getctime(self.cache.root_folder + "/" + self.filename)

        def get(self, max_age: Optional[float] = None) -> Optional[bytes]:
            try:
                if max_age is not None:
                    if self.age() > max_age:  # expired
                        logging.info("expired %s (age: %.f > %.f)", self.key, self.age(), max_age)
                        # TODO remove?
                        return None
                with open(self.cache.root_folder + "/" + self.filename, "rb") as f:
                    return f.read()
            except FileNotFoundError:
                return None

        def touch(self):
            file_path = self.cache.root_folder + "/" + self.filename
            if not os.path.exists(file_path):
                return
            os.utime(file_path)

        def unlink(self):
            try:
                os.unlink(self.cache.root_folder + "/" + self.filename)
            except FileNotFoundError:
                logging.info(
                    "removing but already gone: %s (%s)", self.key, self.cache.root_folder + "/" + self.filename
                )

        def __enter__(self):
            self.lock.acquire()
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.lock.release()

    def __init__(self, folder: str, max_entries=100, max_bytes=64 * 1024 * 1024):
        self.lock = threading.Lock()
        self.cache = []
        self.root_folder = folder
        self.max_entries = max_entries
        self.max_bytes = max_bytes
        self.max_entry_size = max_bytes // 16
        self.tot_bytes = 0

        os.makedirs(self.root_folder, exist_ok=True)
        # list files by mtime
        files = sorted(os.listdir(self.root_folder), key=lambda f: os.path.getmtime(self.root_folder + "/" + f))
        for f in files:
            key = base64.b64decode(f.encode()).decode()
            e = Cache.Entry(key, self)
            size = os.path.getsize(self.root_folder + "/" + f)
            self.tot_bytes += size
            assert f == e.filename
            # logging.info("recovered %s file %s", size2human(size), key)
            self.cache.append(e)

        self._make_space(0)
        logging.info("recovered %d files for a total of %s", len(self.cache), size2human(self.tot_bytes))

    def _make_space(self, space_needed):
        with self.lock:
            while self.tot_bytes + space_needed > self.max_bytes:
                # FIXME: Needs to be properly handled
                if len(self.cache) == 0:
                    self.tot_bytes = 0
                    return
                e = self.cache.pop(0)
                try:
                    size = os.path.getsize(self.root_folder + "/" + e.filename)
                    self.tot_bytes -= size
                    e.unlink()
                    logging.info("evicted %s file %s", size2human(size), e.key)
                except FileNotFoundError:
                    logging.info("evicted but already gone: %s", e.key)

    def remove(self, key):
        with self.lock:
            for e in self.cache:
                if e.key == key:
                    self.cache.remove(e)
                    e.unlink()
                    self.tot_bytes -= e.size

    def key(self, key):
        with self.lock:
            for e in self.cache:
                if e.key == key:
                    self.cache.remove(e)
                    self.cache.append(e)  # move to end
                    return e
            if len(self.cache) >= self.max_entries:
                e = self.cache.pop(0)
                self.tot_bytes -= e.size
                e.unlink()
            e = Cache.Entry(key, self)
            self.cache.append(e)
            return e



import re
import time
from contextlib import contextmanager
from contextvars import ContextVar
from math import log2
from typing import Callable

IEC_UNITS = ["KiB", "MiB", "GiB", "TiB"]


_RE_ID_AND_DOTS = re.compile(r"[a-zA-Z][a-zA-Z0-9_\\.]*")


class FieldNotFound(Exception):
    def __init__(self, field: str):
        super().__init__(f"Field '{field}' not found")


def is_valid_id(s: str) -> bool:
    return _RE_ID_AND_DOTS.fullmatch(s) is not None


def size2human(size: int) -> str:
    if size == 0:
        return "0B"
    p = int(log2(size) // 10.0)

    if p < 1:
        return f"{size}B"
    if p > len(IEC_UNITS):
        p = len(IEC_UNITS)
    converted_size = size / 1024**p
    return f"{converted_size:.1f}{IEC_UNITS[p - 1]}"


_TIMED: ContextVar = ContextVar("timed", default=None)


@contextmanager
def timing(name: str):
    t0 = time.perf_counter()
    old = _TIMED.get()
    if old is None:
        old = lambda k, v: None
    inner = 0.0

    def inc(name, elapsed):
        nonlocal inner
        old(name, elapsed)
        inner += elapsed

    _TIMED.set(inc)
    try:
        yield
    finally:
        _TIMED.set(old)
        elapsed = time.perf_counter() - t0 - inner
        if old is not None:
            old(name, elapsed)


@contextmanager
def start_timing(name: str, cb: Callable[[str, float], None]):
    old = _TIMED.get()
    _TIMED.set(cb)
    try:
        with timing(name):
            yield
    finally:
        _TIMED.set(old)
